// Taken from http://penguin.ewu.edu/~trolfe/CCSC2002/Histogram.html
// Modified Tristan Aubrey-Jones 2013

#include <iostream>
#include <iomanip>    // setwidth, etc.
#include <cstdlib>     // srand, rand, etc.
#include <ctime>       // time for srand's use
#include <cmath>
#include <mpi.h>

#define  BUFSIZE  1     // MPI tag 1
#define  BUFFER   2     // MPI tag 2

/* --- DISABLE "typedef  long  Number;"
       We need to hard-code the data type for MPI
*/

using namespace std;

void Start(int MyPos, int &Nh, int &Size, double* &X, double Lo, double Hi);

void Concentrate (int MyPos, int Nh, int *Hist);

void Pause();

// Fill with random data in the indicated range --- Primus only
void Fill (double *X, int N, double Lo, double Hi);

// Count hits for a histogram
void Histogram ( double *X, int N, int *Hist, int Nh,
                 double Lo, double Hi );

// TIME

#define MAX_TIME_HEADER 1024
int timeRank;
char timeHeader[MAX_TIME_HEADER];
double beginTime, middleTime;

void initTime() {
  // rank
  timeRank = MPI::COMM_WORLD.Get_rank();

  // start time
  beginTime = MPI_Wtime();
  middleTime = beginTime;
}

void logTime(const char* event, double *startTime) {
  if (timeRank == 0) {
    double End = MPI_Wtime();
    std::cerr << timeHeader << "\t" 
              << event << "\t" << (End - (*startTime)) << std::endl;
  }
  middleTime = MPI_Wtime();
}

int main (int argc, char *argv[])
{  double   *Data = NULL; // Dynamic data array
   double    Lo = 0.0,      // Data range information
           Hi = 10000.0;
   int     Size,        // Total array size
           N;           // For Primus, original value of Size
   int    *Hist,        // Counts for the histogram
          *Part,        // Second array (sequential version)
           Nh,          // Number of histogram blocks
           Nproc;       // Number of processes cooperating
   int     MyPos,       // My position MPI_COMM_WORLD
           rc;          // return code from MPI function

   double  Begin,       // Beginning of parallel segment
           Middle,      // End of parallel segment, start of sequential
           Finish;      // End of sequential segment

   int     Primus;
   time_t timer;
   struct tm * dateTime;
   char* strTime, *strPtr;

   rc = MPI_Init(&argc, &argv);
   if (rc != MPI_SUCCESS)
   {  cout << "MPI_Init failed\n"; exit(-1);  }
   
   rc = MPI_Comm_rank (MPI_COMM_WORLD, &MyPos);
   rc = MPI_Comm_size (MPI_COMM_WORLD, &Nproc);

   Primus = (MyPos == 0);

   cout << "Size:  ";
   if ( argc > 1 )
   {
      Size = atoi (argv[1]);
      cout << Size << endl;
   }
   else
   {
     cout << "Must specify arg 1 - size!\n";
     MPI_Finalize();
     exit(-1);
   }
   cout << "N divisions:  ";
   if ( argc > 2 )
   {
      Nh = atoi (argv[2]);
      cout << Nh << endl;
   }
   else
   {
     cout << "Must specify arg 2 - Number divisions!\n";
     MPI_Finalize();
     exit(-1);
   }

   cout << "N Processes:  " << Nproc << endl;

   N = Size;          // Spare copy

   // init time
    // the current date/time
  time_t timer1 = time(NULL);
  struct tm* dateTime1 = localtime(&timer1);
  char *strTime1 = asctime(dateTime1);
  char* strPtr1 = strTime1;
  while (*strPtr1 != '\0') { if (*strPtr1 == '\n') *strPtr1 = '\0'; strPtr1++; }
  
  // make log header
  snprintf(timeHeader, MAX_TIME_HEADER, "%s\t%i\t%i\t%i", strTime1, N, Nh, Nproc);

  // time the computation
  initTime();
   Begin = MPI_Wtime();

   // (Nolonger scatters data from root, but
   // generates data locally at each node)
   Start(MyPos, Nh, Size, Data, Lo, Hi);
   logTime("genfloats", &beginTime);

   cout << MyPos << ": finished creating data" << endl;
   //if ( Primus ) cout << "Finished with clone starts" << endl;

   Hist  = (int*) calloc (Nh, sizeof(int));
   cout << MyPos << ": allocated histogram" << endl;
    
   Histogram(Data, Size, Hist, Nh, Lo, Hi);
   logTime("localparts", &middleTime);
   cout << MyPos << ": computed local histogram" << endl;

   if (Nproc > 1) {
     cout << MyPos << ": concentrating hist at root" << endl;
     Concentrate (MyPos, Nh, Hist);
     cout << MyPos << ": concentrated hist at root" << endl;
   }
   logTime("exchangedparts", &middleTime);

   MPI_Finalize();          // Everybody out of MPI
   Middle = MPI_Wtime();

   if ( !Primus ) exit(0);  // Only Primus keep going.

   logTime("total", &beginTime);

   cout << "Back to just one process." << endl;
   //cout << "Processing pure sequential approach" << endl;

  // alteration, don't do sequental, just end here
// Sequential algorithm --- using the saved value in N
   //Part  = (int*) calloc (Nh, sizeof(int));
   //Histogram(Data, N, Part, Nh, Lo, Hi);

   Finish = MPI_Wtime();

   for ( int Seg = 1; Seg <= Nh; Seg++ )
   {  double Range = 1.0 + Hi - Lo;

      cout << setw(3) << Seg << ":" << setw(9) << Hist[Seg-1]
           //<< setw(9) << Part[Seg-1]
           << " below " << Lo + Range*Seg/Nh << endl;
   }

   cout << setprecision(3)
        << "\nParallel:    " << Middle - Begin
        << "\nSequential:  " << Finish - Middle << endl;

   cout << setprecision(3)
        << "\nParallel:    " << Middle - Begin
        << "\nSequential:  " << Finish - Middle << endl;
   cout << "FINISHED" << endl;

  free(Data);

}

// Fill with random data in the indicated range
void Fill (double *X, int N, double Lo, double Hi)
{  double Range = Hi - Lo;

   if (Hi+0.1 == Hi)
      Range++;

   for (int Idx = 0; Idx < N; Idx++)
      X[Idx] = double(rand())*Range/(RAND_MAX+1.0) + Lo;
}

// Count hits for a histogram.
// Note:  range presumed to be (Hi+1)-Lo.
void Histogram ( double *X, int N, int *Hist, int Nh,
                 double Lo, double Hi )
{  double Range;     // Total range of values
   int  Cell;      // Subscript for histogram cell to increment

   Range = 1 + Hi - Lo;

   while ( N-- )   // Single pass through array, so use pointer
   {//Note:  update the point X on the fly.
      Cell = int(double(*X++ - Lo) / Range * Nh);
      Hist[Cell]++;
   }
}

void Start(int MyPos, int &Nh, int &Size, double* &X, double Lo, double Hi)
{
  // calculate size for this segment
  int        Nproc;
  MPI_Comm_size (MPI_COMM_WORLD, &Nproc);
  int TotalSize = Size;
  Size /= Nproc; 
  if (MyPos < TotalSize % Nproc) Size++;

  // generate data locally
  X  = (double*) calloc (Size, sizeof(double));
  if ( !X )
  {  cout << "Insufficient space.\n"; exit(-3);  }

  srand (time(NULL));
  Fill (X, Size, Lo, Hi);

  return;
  // DONT CONTINUE BELOW.

   MPI_Status Status;

   if ( MyPos == 0 )      // Initiating process
   {  int        Front;   // For walking through in the split
      int        Pending; // Number still unprocessed
      int        Seg,     // Loop variable:  segment being worked
                 Buf[2] = { Nh },
                 SegSize; // Elements in a segment
      int        k;
      int        MySize;

      MPI_Comm_size (MPI_COMM_WORLD, &Nproc);

      MySize = Size / Nproc;
      if ( Size % Nproc ) MySize++;
      Front = MySize;
      Pending = Size - MySize;

      for ( Seg = Nproc-1, k = 1;
            Seg > 0; Seg--, k++ )
      {  SegSize = Pending / Seg;

         if ( Pending % Seg ) ++SegSize;
         Buf[1] = SegSize;

         MPI_Send (Buf, 2, MPI_INT, k, BUFSIZE, MPI_COMM_WORLD);
         MPI_Send (X+Front, SegSize, MPI_DOUBLE, k, BUFFER, MPI_COMM_WORLD);

         Front   += SegSize;
         Pending -= SegSize;
      }
      Size = MySize;     // Modify Size for Primus itself
   }
   else                           // Child process
   {
      int Buf[2];

      MPI_Recv(Buf, 2, MPI_INT, 0, BUFSIZE, MPI_COMM_WORLD, &Status);
      Nh = Buf[0];  Size = Buf[1];
      X = (double*) calloc (Size, sizeof(double));
      MPI_Recv(X, Size, MPI_DOUBLE, 0, BUFFER, MPI_COMM_WORLD, &Status);
    }
}

void Concentrate (int MyPos, int Nh, int *Hist)
{
   int *RecvBuf = (int*) calloc (Nh, sizeof(int)); // was sizeof(long)

   cout << MyPos << ": before barrier" << endl;
   MPI_Barrier(MPI_COMM_WORLD);

   cout << MyPos << ": after barrier, before reduce, Nh=" << Nh << endl;
   MPI_Reduce ( Hist, RecvBuf, Nh, MPI_INT, MPI_SUM, 0,
                MPI_COMM_WORLD );
   cout << MyPos << ": after reduce" << endl;

   if ( !MyPos )
      memcpy ( Hist, RecvBuf, Nh * sizeof(*Hist) );
   free(RecvBuf);
   cout << MyPos << ": finished concentrate" << endl;
}
